/*
 * trafficcontr.c
 * 
 * created by: Timo Lindhorst <timo.lindhorst@student.uni-magdeburg.de>
 * 
 * listen to a port and throttle down traffic if a transmission is detected.
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <getopt.h>
#include <signal.h>

#define RECV_BUFFER_SIZE 1500
#define SEC_TO_WAIT 3
#define USEC_TO_WAIT 0

static void throttle(int state);
static int read_packet(int udpSocket, int *buffer);

char cmdThrottle[100];
char cmdRaise[100];
int throttling = 0;

void quit(int sig) {
    if (throttling == 1) { 
        system(cmdRaise);
        printf("not throttling\n");
    }
    exit(0);
}
    
int main (int argc, char **argv) {
	int udpSocket;
    struct sockaddr_in receiverAddr;
	int buffer[RECV_BUFFER_SIZE/sizeof(int)];
	int broadcast;
	fd_set rfds;
	struct timeval tv;   
    int port = 55555;
    int msecsDelay = 3000;
    int rate = 800;
    char* interface = "wlan0";
        
    extern char *optarg;
    
    struct option long_options[] = {
        { "port", required_argument, 0, 'p' },
        { "delay", required_argument, 0, 'd' },
        { "rate", required_argument, 0, 'r' },
        { "interface", required_argument, 0, 'i' },
        { "help", no_argument, 0, 'h' }
    };
    
    int option;
    
    signal(SIGINT, quit);
    
    while((option = getopt_long(argc, argv, "p:d:r:i:h", long_options, 0)) != -1) {
        switch(option) {
            case 'p':
                port = atoi(optarg);
                break;
            case 'd':
                msecsDelay = atoi(optarg);
                break;
            case 'r':
                rate = atoi(optarg);
                break;
            case 'i':
                interface = optarg;
                break;
            case 'h':
            default:
                printf("use: trafficcontrol [option]\n\n");
                printf("Options: \n");
                printf("\t-p, --port=PORT          send to port PORT\n");
                printf("\t-d, --delay=DELAY        wait DELAY msecs before raising rate again\n");
                printf("\t-r, --rate=RATE          throttle the rate to RATE kbit/s\n");
                printf("\t-i, --interface=IF       throttle interface IF\n");
                printf("\t-h, --help               show this help\n");
                exit(1);
        }
    }
        
    printf("trafficcontr - listening on port %i, throttle interface %s to %i kbit/s, raise after %i msecs\n", port, interface, rate, msecsDelay);
    
    sprintf(cmdThrottle, "tc qdisc add dev %s root tbf rate %ikbit latency 100ms burst 2000",
            interface, rate);
    sprintf(cmdRaise, "tc qdisc del dev %s root", interface);
                
	udpSocket = socket(PF_INET, SOCK_DGRAM, 0);
	if (udpSocket < 0) {
		perror("error opening socket!");
		exit(1);
	}
	
	setsockopt(udpSocket, SOL_SOCKET, SO_BROADCAST, &broadcast, sizeof(broadcast));
	
	receiverAddr.sin_family = AF_INET;
	receiverAddr.sin_port   = htons(port);
	receiverAddr.sin_addr.s_addr = INADDR_ANY;
		
	bind(udpSocket, (struct sockaddr*) &receiverAddr, sizeof(receiverAddr));
	    	
	while(1) {
        tv.tv_sec = msecsDelay / 1000;
        tv.tv_usec = (msecsDelay % 1000) * 1000;
                    
          FD_ZERO(&rfds);
          FD_SET(udpSocket, &rfds);
                          
          if(select(udpSocket + 1, &rfds, NULL, NULL, &tv) == 0) {
            throttle(0);
          } else if (FD_ISSET(udpSocket, &rfds)) {
            if(read_packet(udpSocket, buffer) < 0) {
              printf("read failed!\n");
            } else {
              throttle(1);
            }
          }
	}
	close(udpSocket);
	return(0);
}

static void throttle(int state) {
    if (throttling == state) return;
    
    switch(state) {
      case 0:
        system(cmdRaise);
        printf("not throttling\n");
        break;
      case 1:
        system(cmdThrottle);
        printf("throttling\n");
        break;
    }
    throttling = state;
}

static int read_packet(int udpSocket, int *buffer) {
    return(recv(udpSocket, buffer, RECV_BUFFER_SIZE, 0));
}
